# Pipfile & Pipfile.lock

Pipenv uses two files to manage project dependencies: `Pipfile` and `Pipfile.lock`. This document explains their purpose, structure, and best practices for working with them.

## Overview

| File | Purpose | Managed By | Format |
|------|---------|------------|--------|
| `Pipfile` | Declares top-level project dependencies and constraints | Developers | [TOML](https://toml.io/en/latest) |
| `Pipfile.lock` | Records exact versions and hashes of all resolved dependencies | Pipenv automatically | JSON |

Both files should be committed to version control to ensure consistent environments across development and deployment.

## Pipfile

The `Pipfile` is a human-readable, TOML-formatted file that declares your project's dependencies. It replaces the traditional `requirements.txt` file with a more powerful and flexible format.

### Pipfile Structure

A typical `Pipfile` contains the following sections:

```toml
[[source]]
# Package sources (PyPI, private repositories, etc.)

[packages]
# Production dependencies

[dev-packages]
# Development dependencies

[requires]
# Python version requirements

[scripts]
# Custom script definitions

[pipenv]
# Pipenv configuration directives

[custom-category]
# Custom package categories (e.g., docs, tests)
```

### Source Section

The `[[source]]` section defines where packages should be downloaded from:

```toml
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

# You can specify multiple sources
[[source]]
url = "https://private-repo.example.com/simple"
verify_ssl = true
name = "private"
```

### Packages Section

The `[packages]` section lists production dependencies:

```toml
[packages]
# Simple version specification
requests = "*"                # Any version
flask = "==2.0.1"             # Exact version
numpy = ">=1.20.0,<2.0.0"     # Version range
pandas = "~=1.3.0"            # Compatible release

# Extended syntax with additional options
django = {version = ">=3.2", extras = ["bcrypt"]}
sentry-sdk = {version = ">=1.0.0", extras = ["flask"]}

# Git repositories
flask-login = {git = "https://github.com/maxcountryman/flask-login.git", ref = "master"}

# Local paths
my-package = {path = "./path/to/local/package", editable = true}

# Platform-specific dependencies
gunicorn = {version = "*", markers = "sys_platform == 'linux'"}
waitress = {version = "*", markers = "sys_platform == 'win32'"}

# Index-specific packages
private-package = {version = "*", index = "private"}
```

### Development Packages Section

The `[dev-packages]` section lists dependencies needed only for development:

```toml
[dev-packages]
pytest = ">=6.0.0"
black = "==21.5b2"
mypy = "*"
sphinx = {version = ">=4.0.0", extras = ["docs"]}
```

### Python Version Requirements

The `[requires]` section specifies Python version constraints:

```toml
[requires]
python_version = "3.9"        # Requires Python 3.9.x
# OR
python_full_version = "3.9.6" # Requires exactly Python 3.9.6
```

### Custom Scripts

The `[scripts]` section defines shortcuts for common commands:

```toml
[scripts]
start = "python app.py"
test = "pytest tests/"
lint = "flake8 ."
docs = "sphinx-build -b html docs/ docs/_build"
```

You can run these scripts using `pipenv run <script-name>`.

### Pipenv Directives

The `[pipenv]` section controls Pipenv's behavior:

```toml
[pipenv]
allow_prereleases = true       # Allow pre-release versions
disable_pip_input = true       # Prevent pipenv from asking for input
install_search_all_sources = true  # Search all sources when installing from lock
sort_pipfile = true            # Sort packages alphabetically
```

### Custom Package Categories

You can define custom package categories beyond the standard `packages` and `dev-packages`:

```toml
[docs]
sphinx = ">=4.0.0"
sphinx-rtd-theme = "*"

[tests]
pytest = "*"
pytest-cov = "*"
```

## Pipfile.lock

The `Pipfile.lock` file is automatically generated by Pipenv when you run `pipenv lock` or when you install/uninstall packages. It contains:

1. Exact versions of all direct and transitive dependencies
2. Cryptographic hashes for each package
3. Package metadata and requirements

### Pipfile.lock Structure

```json
{
    "_meta": {
        "hash": {
            "sha256": "<hash-of-pipfile-contents>"
        },
        "pipfile-spec": 6,
        "requires": {
            "python_version": "3.9"
        },
        "sources": [
            {
                "name": "pypi",
                "url": "https://pypi.org/simple",
                "verify_ssl": true
            }
        ]
    },
    "default": {
        // Production dependencies and their sub-dependencies
        "package-name": {
            "hashes": [
                "sha256:hash1",
                "sha256:hash2"
            ],
            "index": "pypi",
            "version": "==1.2.3"
        },
        // ...
    },
    "develop": {
        // Development dependencies and their sub-dependencies
        // ...
    },
    "docs": {
        // Custom category dependencies
        // ...
    }
}
```

### Package Categories in Pipfile.lock

Note the naming difference between Pipfile and Pipfile.lock:

- `[packages]` in Pipfile corresponds to `"default"` in Pipfile.lock
- `[dev-packages]` in Pipfile corresponds to `"develop"` in Pipfile.lock
- Custom categories use the same name in both files

## Best Practices

### Version Specifiers

Choose appropriate version specifiers based on your needs:

| Specifier | Example | Meaning |
|-----------|---------|---------|
| `*` | `requests = "*"` | Any version (not recommended for production) |
| `==` | `flask = "==2.0.1"` | Exact version |
| `>=` | `django = ">=3.2"` | Minimum version |
| `<=` | `numpy = "<=1.20.0"` | Maximum version |
| `>=,<` | `pandas = ">=1.3.0,<2.0.0"` | Version range |
| `~=` | `pytest = "~=6.2.0"` | Compatible release (equivalent to `>=6.2.0,<6.3.0`) |

For production environments, it's recommended to use specific version constraints to ensure reproducibility.

### Dependency Management Workflow

1. **Initial setup**: Create a `Pipfile` with your top-level dependencies
   ```bash
   $ pipenv install requests flask
   $ pipenv install pytest --dev
   ```

2. **Lock dependencies**: Generate a `Pipfile.lock` with exact versions
   ```bash
   $ pipenv lock
   ```

3. **Install from lock**: Install the exact versions from `Pipfile.lock`
   ```bash
   $ pipenv sync
   ```

4. **Update dependencies**: Update to newer versions when needed
   ```bash
   $ pipenv update
   ```

### Importing from requirements.txt

If you're migrating from a project that uses `requirements.txt`, you can import it:

```bash
$ pipenv install -r requirements.txt
```

This will create a `Pipfile` and install all packages from the requirements file.

### Generating requirements.txt

You can generate a `requirements.txt` file from your `Pipfile.lock`:

```bash
$ pipenv requirements > requirements.txt
```

This is useful for environments that don't support Pipenv directly.

### Security Considerations

The `Pipfile.lock` includes cryptographic hashes for each package, which are verified during installation. This prevents supply chain attacks where a malicious package could be substituted.

For CI/CD deployments:

1. Use `pipenv verify` to ensure the lock file is up-to-date
2. Use `pipenv install --deploy` to fail if the lock file is out of sync
3. Never run commands that modify the lock file in CI/CD (like `lock`, `update`, or `upgrade`)

## Advanced Usage

### Package Markers

You can use [PEP 508](https://www.python.org/dev/peps/pep-0508/) markers to specify environment-specific dependencies:

```toml
[packages]
gunicorn = {version = "*", markers = "sys_platform == 'linux'"}
waitress = {version = "*", markers = "sys_platform == 'win32'"}
colorama = {version = "*", markers = "python_version >= '3.7'"}
```

### Package Extras

Many packages provide optional features as "extras":

```toml
[packages]
requests = {version = "*", extras = ["socks", "security"]}
django = {version = "*", extras = ["bcrypt"]}
```

### Git Dependencies

You can install packages directly from Git repositories:

```toml
[packages]
flask-login = {git = "https://github.com/maxcountryman/flask-login.git", ref = "master"}
custom-package = {git = "https://github.com/user/repo.git", editable = true}
```

The `ref` parameter can be a branch name, tag, or commit hash.

### Local Dependencies

For local development of packages:

```toml
[packages]
my-package = {path = "./path/to/package", editable = true}
```

The `editable` flag installs the package in development mode, so changes to the source code are immediately reflected.

## Troubleshooting

### Lock File Hash Mismatch

If you see an error about the Pipfile.lock hash not matching:

```
Pipfile.lock out of date, update it with "pipenv lock" or "pipenv update".
```

This means your `Pipfile` has been modified since the last time `Pipfile.lock` was generated. Run `pipenv lock` to update the lock file.

### Dependency Resolution Failures

If Pipenv can't resolve dependencies:

1. Try clearing the cache: `pipenv lock --clear`
2. Check for conflicting version requirements
3. Consider relaxing version constraints in your Pipfile
4. Use `pipenv install --verbose` to see detailed resolution information

### Manually Editing Files

- **Pipfile**: Safe to edit manually, but follow TOML syntax rules
- **Pipfile.lock**: Do not edit manually; always use Pipenv commands to modify
